@@ -68,6 +68,8 @@ class Testbed:
         # used for tracking kernel version changes
         self.last_test_name = ''
         self.last_reboot_marker = ''
+        self.eatmydata_prefix = []
+        self.apt_pin_for_pockets = []
 
         adtlog.debug('testbed init')
 
@@ -188,6 +190,12 @@ class Testbed:
         self.dpkg_arch = self.check_exec(['dpkg', '--print-architecture'], True).strip()
         adtlog.info('testbed dpkg architecture: ' + self.dpkg_arch)
 
+        # do we have eatmydata?
+        (code, out, err) = self.execute(['which', 'eatmydata'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        if code == 0:
+            adtlog.debug('testbed has eatmydata')
+            self.eatmydata_prefix = [out.strip()]
+
         # record package versions of pristine testbed
         if self.output_dir and self.execute(['which', 'dpkg-query'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)[0] == 0:
             pkglist = TempPath(self, 'testbed-packages', autoclean=False)
@@ -223,10 +231,10 @@ class Testbed:
             adtlog.debug('Copying file %s to testbed %s' % (host, tb))
             Path(self, host, tb, os.path.isdir(host)).copydown()
 
-        for p in self.add_apt_pockets:
-            script = '''awk '/^deb(-src)? .*(ubuntu.com|debian.org|ftpmaster)/ { if ($3 !~ /-/) { $3 = $3"-%s"; print }}' ''' \
-                '''/etc/apt/sources.list `ls /etc/apt/sources.list.d/*.list 2>/dev/null|| true` ''' \
-                ''' > /etc/apt/sources.list.d/%s.list''' % (p, p)
+        # create apt sources for --apt-pocket
+        for pocket in self.add_apt_pockets:
+            pocket = pocket.split('=', 1)[0]  # strip off package list
+            script = '''sed -rn 's/^(deb|deb-src) +(\[.*\] *)?([^ ]*(ubuntu.com|debian.org|ftpmaster|file:\/\/\/tmp\/adttestarchive)[^ ]*) +([^ -]+) +(.*)$/\\1 \\2\\3 \\5-%s \\6/p' /etc/apt/sources.list `ls /etc/apt/sources.list.d/*.list 2>/dev/null|| true)` > /etc/apt/sources.list.d/%s.list; ''' % (pocket, pocket)
             self.check_exec(['sh', '-ec', script])
 
         # record the mtimes of dirs affecting the boot
@@ -244,6 +252,15 @@ class Testbed:
             if rc:
                 self.bomb('testbed setup commands failed with status %i' % rc)
 
+        # create apt pinning for --apt-pocket with package list
+        for pocket in self.add_apt_pockets:
+            # do we have a package list?
+            try:
+                (pocket, pkglist) = pocket.split('=', 1)
+            except ValueError:
+                continue
+            self._create_apt_pinning_for_packages(pocket, pkglist)
+
         if self.setup_commands and 'reboot' in self.caps:
             boot_affected = self.execute(
                 ['bash', '-ec', '[ ! -e /run/autopkgtest_no_reboot.stamp ] || exit 0;'
@@ -448,24 +465,36 @@ Description: satisfy autopkgtest test dependencies
         shutil.rmtree(pkgdir)
         deb.copydown()
 
-        # install it and its dependencies in the tb
-        self.check_exec(['dpkg', '--unpack', deb.tb], stdout=subprocess.PIPE)
-        rc = self.execute(['apt-get', 'install', '--quiet', '--quiet', '--assume-yes', '--fix-broken',
-                           '-o', 'APT::Install-Recommends=%s' % recommends,
-                           '-o', 'Debug::pkgProblemResolver=true'],
-                          kind='install')[0]
-        if rc != 0:
-            if shell_on_failure:
-                self.run_shell()
-            self.badpkg('failed to run apt-get to satisfy adt-satdep.deb dependencies')
-        rc = self.execute(['dpkg', '--status', 'adt-satdep'],
-                          stdout=subprocess.PIPE,
-                          stderr=subprocess.PIPE)[0]
-        if rc != 0:
-            self.badpkg('Test dependencies are unsatisfiable. A common reason is '
-                        'that your testbed is out of date with respect to the '
-                        'archive, and you need to use a current testbed or run '
-                        'apt-get update or use -U.')
+        # install it and its dependencies in the tb; our apt pinning is not
+        # very clever wrt. resolving transitional dependencies in the pocket,
+        # so we might need to retry without pinning
+        while True:
+            self.check_exec(['dpkg', '--unpack', deb.tb], stdout=subprocess.PIPE)
+            rc = self.execute(self.eatmydata_prefix +
+                              ['apt-get', 'install', '--assume-yes', '--fix-broken',
+                               '-o', 'APT::Install-Recommends=%s' % recommends,
+                               '-o', 'Debug::pkgProblemResolver=true'],
+                              kind='install')[0]
+            if rc != 0:
+                if shell_on_failure:
+                    self.run_shell()
+                self.badpkg('failed to run apt-get to satisfy adt-satdep.deb dependencies')
+            rc = self.execute(['dpkg', '--status', 'adt-satdep'],
+                              stdout=subprocess.PIPE,
+                              stderr=subprocess.PIPE)[0]
+            if rc != 0:
+                if self.apt_pin_for_pockets:
+                    pocket = self.apt_pin_for_pockets.pop()
+                    adtlog.warning('Test dependencies are unsatisfiable with using apt pinning. '
+                                   'Retrying with using all packages from %s' % pocket)
+                    self.check_exec(['/bin/sh', '-ec', 'rm /etc/apt/preferences.d/autopkgtest-*-' + pocket])
+                    continue
+
+                self.badpkg('Test dependencies are unsatisfiable. A common reason is '
+                            'that your testbed is out of date with respect to the '
+                            'archive, and you need to use a current testbed or run '
+                            'apt-get update or use -U.')
+            break
 
         # remove adt-satdep to avoid confusing tests, but avoid marking our
         # test dependencies for auto-removal
@@ -857,6 +886,7 @@ fi
                  'export ADTTMP=$(mktemp -d --tmpdir adttmp.XXXXXX); ' \
                  'export DEBIAN_FRONTEND=noninteractive; ' \
                  'export LANG=C.UTF-8; ' \
+                 '''export DEB_BUILD_OPTIONS=parallel=$(grep -c ^processor /proc/cpuinfo | sed 's/^0$/1/'); ''' \
                  'unset LANGUAGE LC_CTYPE LC_NUMERIC LC_TIME LC_COLLATE '\
                  '  LC_MONETARY LC_MESSAGES LC_PAPER LC_NAME LC_ADDRESS '\
                  '  LC_TELEPHONE LC_MEASUREMENT LC_IDENTIFICATION LC_ALL;' \
@@ -1026,6 +1056,45 @@ fi
         else:
             adtlog.debug('no need to restore click AppArmor profiles')
 
+    #
+    # helper methods
+    #
+
+    def _create_apt_pinning_for_packages(self, pocket, pkglist):
+        '''Create apt pinning for --apt-pocket=pocket=pkglist'''
+
+        # sort pkglist into source and binary packages
+        binpkgs = []
+        srcpkgs = []
+        for i in pkglist.split(','):
+            i = i.strip()
+            if i.startswith('src:'):
+                srcpkgs.append(i[4:])
+            else:
+                binpkgs.append(i)
+
+        # get release name
+        script = 'SRCS=$(ls /etc/apt/sources.list /etc/apt/sources.list.d/*.list 2>/dev/null|| true); '
+        script += '''REL=$(sed -rn '/^(deb|deb-src) .*(ubuntu.com|debian.org|ftpmaster|file:\/\/\/tmp\/adttestarchive)/ { s/^[^ ]+ +(\[.*\] *)?[^ ]* +([^ -]+) +.*$/\\2/p}' $SRCS | head -n1); '''
+
+        script += 'mkdir -p /etc/apt/preferences.d; '
+        script += 'PKGS="%s"; ' % ' '.join(binpkgs)
+
+        # translate src:name entries into binaries of that source
+        if srcpkgs:
+            script += 'PKGS="$PKGS $(apt-cache showsrc %s | ' \
+                '''sed -n '/^Binary: / {s/^Binary: //; s/,//g; p}' | ''' \
+                'xargs -n1 | sort -u | xargs)"; ' % \
+                ' '.join(srcpkgs)
+
+        # prefer given packages from pocket, other packages from
+        # default $REL (prio 900), but make $REL-pocket available for
+        # dependency resolution (prio 800)
+        script += 'printf "Package: $PKGS\\nPin: release a=${REL}-%(pocket)s\\nPin-Priority: 990\\n\\nPackage: *\\nPin: release a=$REL\\nPin-Priority: 900\\n\\nPackage: *\\nPin: release a=${REL}-%(pocket)s\\nPin-Priority: 800\\n" > /etc/apt/preferences.d/autopkgtest-${REL}-%(pocket)s; ' % \
+            {'pocket': pocket}
+        self.check_exec(['sh', '-ec', script])
+        self.apt_pin_for_pockets.append(pocket)
+
 
 class Path:
     '''Represent a file/dir with a host and a testbed path'''
